@@ -12,13 +12,14 @@ Python libraries, though it does make heavy use of external packaging commands.
 Visit <https://github.com/datalad/datalad-installer> for more information.
 """
 
-__version__ = "0.3.1"
+__version__ = "0.5.0"
 __author__ = "The DataLad Team and Contributors"
 __author_email__ = "team@datalad.org"
 __license__ = "MIT"
 __url__ = "https://github.com/datalad/datalad-installer"
 
 from abc import ABC, abstractmethod
+from contextlib import contextmanager
 import ctypes
 from enum import Enum
 from functools import total_ordering
@@ -42,6 +43,7 @@ from typing import (
     Callable,
     ClassVar,
     Dict,
+    Iterator,
     List,
     NamedTuple,
     Optional,
@@ -702,10 +704,11 @@ class DataladInstaller:
         Add a line to the env write files that prepends (or appends, if
         ``last`` is true) a given path to ``PATH``
         """
+        path = Path(p).resolve()
         if not last:
-            line = f'export PATH={shlex.quote(str(p))}:"$PATH"'
+            line = f'export PATH={shlex.quote(str(path))}:"$PATH"'
         else:
-            line = f'export PATH="$PATH":{shlex.quote(str(p))}'
+            line = f'export PATH="$PATH":{shlex.quote(str(path))}'
         self.addenv(line)
 
     def addcomponent(self, name: str, **kwargs: Any) -> None:
@@ -1415,14 +1418,17 @@ class PipInstaller(Installer):
                 file=script,
                 flush=True,
             )
+            # We need to close before passing to Python for Windows
+            # compatibility
             script.close()
             binpath = Path(readcmd(self.python, script.name))
+            os.unlink(script.name)
         log.debug("Installed program directory: %s", binpath)
         return binpath
 
     def assert_supported_system(self) -> None:
         ### TODO: Detect whether pip is installed in the current Python,
-        ### preferrably without importing it
+        ### preferably without importing it
         pass
 
 
@@ -1451,6 +1457,12 @@ class DebURLInstaller(Installer):
 
     OPTIONS = [
         Option("--url", metavar="URL", help="URL from which to download `*.deb` file"),
+        Option(
+            "--install-dir",
+            converter=Path,
+            metavar="DIR",
+            help="Directory in which to unpack the `*.deb`",
+        ),
         EXTRA_ARGS_OPTION,
     ]
 
@@ -1463,6 +1475,7 @@ class DebURLInstaller(Installer):
         self,
         package: str,
         url: Optional[str] = None,
+        install_dir: Optional[Path] = None,
         extra_args: Optional[List[str]] = None,
         **kwargs: Any,
     ) -> Path:
@@ -1470,19 +1483,32 @@ class DebURLInstaller(Installer):
         if url is None:
             raise RuntimeError("deb-url method requires URL")
         log.info("URL: %s", url)
+        if install_dir is not None:
+            if package != "git-annex":
+                raise RuntimeError("--install-dir is only supported for git-annex")
+            install_dir = untmppath(install_dir)
+            log.info("Install dir: %s", install_dir)
         log.info("Extra args: %s", extra_args)
         if kwargs:
             log.warning("Ignoring extra installer arguments: %r", kwargs)
         with tempfile.TemporaryDirectory() as tmpdir:
             debpath = os.path.join(tmpdir, f"{package}.deb")
             download_file(url, debpath)
-            cmd = ["dpkg", "-i"]
-            if extra_args is not None:
-                cmd.extend(extra_args)
-            cmd.append(debpath)
-            self.manager.sudo(*cmd)
-            log.debug("Installed program directory: /usr/bin")
-            return Path("/usr/bin")
+            if install_dir is not None and "{version}" in str(install_dir):
+                deb_version = readcmd(
+                    "dpkg-deb", "--showformat", "${Version}", "-W", debpath
+                )
+                install_dir = Path(str(install_dir).format(version=deb_version))
+                log.info("Expanded install dir to %s", install_dir)
+            binpath = install_deb(
+                debpath,
+                self.manager,
+                Path("usr/bin"),
+                install_dir=install_dir,
+                extra_args=extra_args,
+            )
+            log.debug("Installed program directory: %s", binpath)
+            return binpath
 
     def assert_supported_system(self) -> None:
         if shutil.which("dpkg") is None:
@@ -1639,32 +1665,50 @@ class DataladGitAnnexBuildInstaller(Installer):
     datalad/git-annex
     """
 
-    NAME = "datalad/git-annex"
+    NAME = "datalad/git-annex:tested"
 
-    OPTIONS: ClassVar[List[Option]] = []
+    OPTIONS = [
+        Option(
+            "--install-dir",
+            converter=Path,
+            metavar="DIR",
+            help="Directory in which to unpack the `*.deb`",
+        ),
+    ]
 
     PACKAGES = {
         "git-annex": ("git-annex", ["git-annex"]),
     }
 
-    def install_package(self, package: str, **kwargs: Any) -> Path:
-        log.info("Installing %s via datalad/git-annex", package)
+    def install_package(
+        self, package: str, install_dir: Optional[Path] = None, **kwargs: Any
+    ) -> Path:
+        log.info("Installing %s via %s", package, self.NAME)
+        if install_dir is not None:
+            if not ON_LINUX:
+                raise RuntimeError("--install-dir is only supported on Linux")
+            install_dir = untmppath(install_dir)
+            log.info("Install dir: %s", install_dir)
         if kwargs:
             log.warning("Ignoring extra installer arguments: %r", kwargs)
         assert package == "git-annex"
         with tempfile.TemporaryDirectory() as tmpdir_:
             tmpdir = Path(tmpdir_)
             if ON_LINUX:
-                self.download_latest_git_annex("ubuntu", tmpdir)
+                self.download("ubuntu", tmpdir)
                 (debpath,) = tmpdir.glob("*.deb")
-                self.manager.sudo("dpkg", "-i", debpath)
-                binpath = Path("/usr/bin")
+                binpath = install_deb(
+                    debpath,
+                    self.manager,
+                    Path("usr", "bin"),
+                    install_dir=install_dir,
+                )
             elif ON_MACOS:
-                self.download_latest_git_annex("macos", tmpdir)
+                self.download("macos", tmpdir)
                 (dmgpath,) = tmpdir.glob("*.dmg")
                 binpath = install_git_annex_dmg(dmgpath, self.manager)
             elif ON_WINDOWS:
-                self.download_latest_git_annex("windows", tmpdir)
+                self.download("windows", tmpdir)
                 (exepath,) = tmpdir.glob("*.exe")
                 runcmd(exepath, "/S")
                 binpath = Path("C:/Program Files", "Git", "usr", "bin")
@@ -1681,14 +1725,38 @@ class DataladGitAnnexBuildInstaller(Installer):
             raise MethodNotSupportedError(f"{SYSTEM} OS not supported")
 
     @staticmethod
-    def download_latest_git_annex(ostype: str, target_dir: Path) -> None:
+    def download(ostype: str, target_dir: Path) -> None:
         """
         Download & unzip the artifact from the latest successful build of
         datalad/git-annex for the given OS in the given directory
         """
-        repo = "datalad/git-annex"
-        branch = "master"
-        workflow = f"build-{ostype}.yaml"
+        GitHubArtifactDownloader().download_last_successful_artifact(
+            target_dir, repo="datalad/git-annex", workflow=f"build-{ostype}.yaml"
+        )
+
+
+@GitAnnexComponent.register_installer
+class DataladGitAnnexLatestBuildInstaller(DataladGitAnnexBuildInstaller):
+    """
+    Installs git-annex via the artifact from the latest artifact-producing
+    build (successful or unsuccessful) of datalad/git-annex
+    """
+
+    NAME = "datalad/git-annex"
+
+    @staticmethod
+    def download(ostype: str, target_dir: Path) -> None:
+        """
+        Download & unzip the artifact from the latest build of
+        datalad/git-annex for the given OS in the given directory
+        """
+        GitHubArtifactDownloader().download_latest_artifact(
+            target_dir, repo="datalad/git-annex", workflow=f"build-{ostype}.yaml"
+        )
+
+
+class GitHubArtifactDownloader:
+    def __init__(self) -> None:
         token = os.environ.get("GITHUB_TOKEN")
         if not token:
             r = subprocess.run(
@@ -1702,44 +1770,110 @@ class DataladGitAnnexBuildInstaller(Installer):
                     " environment variable or hub.oauthtoken Git config option."
                 )
             token = r.stdout.strip()
+        self.token: str = token
 
-        def apicall(url: str) -> Any:
-            log.debug("HTTP request: GET %s", url)
-            req = Request(url, headers={"Authorization": f"Bearer {token}"})
-            with urlopen(req) as r:
-                return json.load(r)
+    @contextmanager
+    def get(self, url: str) -> Iterator[Any]:
+        log.debug("HTTP request: GET %s", url)
+        req = Request(url, headers={"Authorization": f"Bearer {self.token}"})
+        with urlopen(req) as r:
+            yield r
 
-        jobs_url = (
-            f"https://api.github.com/repos/{repo}/actions/workflows/{workflow}"
-            f"/runs?status=success&branch={branch}"
-        )
-        log.info("Getting artifacts_url from %s", jobs_url)
-        jobs = apicall(jobs_url)
-        try:
-            artifacts_url = jobs["workflow_runs"][0]["artifacts_url"]
-        except LookupError:
-            log.exception("Unable to get artifacts_url")
-            raise
+    def getjson(self, url: str) -> Any:
+        with self.get(url) as r:
+            return json.load(r)
+
+    def get_workflow_runs(self, url: str) -> Iterator[dict]:
+        while True:
+            with self.get(url) as r:
+                data = json.load(r)
+                for run in data["workflow_runs"]:
+                    assert isinstance(run, dict)
+                    yield run
+                links = parse_header_links(r.headers.get("Link"))
+                url2 = links.get("next", {}).get("url")
+                if url2 is None:
+                    break
+                url = url2
+
+    def get_archive_download_url(self, artifacts_url: str) -> Optional[str]:
+        """
+        Given a workflow run's ``artifacts_url``, returns the
+        ``archive_download_url`` for the one & only artifact.  If there are no
+        artifacts, `None` is returned.  If there is more than one artifact, a
+        `RuntimeError` is raised.
+        """
         log.info("Getting archive download URL from %s", artifacts_url)
-        artifacts = apicall(artifacts_url)
+        artifacts = self.getjson(artifacts_url)
         if artifacts["total_count"] < 1:
-            raise RuntimeError("No artifacts found!")
+            log.debug("No artifacts found")
+            return None
         elif artifacts["total_count"] > 1:
             raise RuntimeError("Too many artifacts found!")
         else:
-            archive_download_url = artifacts["artifacts"][0]["archive_download_url"]
+            url = artifacts["artifacts"][0]["archive_download_url"]
+            assert isinstance(url, str)
+            return url
+
+    def download_archive(self, target_dir: Path, archive_download_url: str) -> None:
+        """
+        Downloads the workflow build artifact zip from ``archive_download_url``
+        and expands it in ``target_dir``
+        """
         log.info("Downloading artifact package from %s", archive_download_url)
         target_dir.mkdir(parents=True, exist_ok=True)
         artifact_path = target_dir / ".artifact.zip"
         download_file(
             archive_download_url,
             artifact_path,
-            headers={"Authorization": f"Bearer {token}"},
+            headers={"Authorization": f"Bearer {self.token}"},
         )
         with ZipFile(str(artifact_path)) as zipf:
             zipf.extractall(str(target_dir))
         artifact_path.unlink()
 
+    def download_latest_artifact(
+        self, target_dir: Path, repo: str, workflow: str, branch: str = "master"
+    ) -> None:
+        """
+        Downloads the most recent artifact built by ``workflow`` on ``branch``
+        in ``repo`` to ``target_dir``
+        """
+        runs_url = (
+            f"https://api.github.com/repos/{repo}/actions/workflows/{workflow}"
+            f"/runs?branch={branch}"
+        )
+        log.info("Getting artifacts_url from %s", runs_url)
+        for run in self.get_workflow_runs(runs_url):
+            artifacts_url = run["artifacts_url"]
+            archive_download_url = self.get_archive_download_url(artifacts_url)
+            if archive_download_url is not None:
+                self.download_archive(target_dir, archive_download_url)
+                return
+        else:
+            raise RuntimeError("No workflow runs with artifacts found!")
+
+    def download_last_successful_artifact(
+        self, target_dir: Path, repo: str, workflow: str, branch: str = "master"
+    ) -> None:
+        """
+        Downloads the most recent artifact built by a succesful run of
+        ``workflow`` on ``branch`` in ``repo`` to ``target_dir``
+        """
+        runs_url = (
+            f"https://api.github.com/repos/{repo}/actions/workflows/{workflow}"
+            f"/runs?status=success&branch={branch}"
+        )
+        log.info("Getting artifacts_url from %s", runs_url)
+        for run in self.get_workflow_runs(runs_url):
+            artifacts_url = run["artifacts_url"]
+            archive_download_url = self.get_archive_download_url(artifacts_url)
+            if archive_download_url is not None:
+                self.download_archive(target_dir, archive_download_url)
+                return
+        else:
+            raise RuntimeError("No workflow runs with artifacts found!")
+
 
 @GitAnnexComponent.register_installer
 class DataladPackagesBuildInstaller(Installer):
@@ -1798,6 +1932,46 @@ class DataladPackagesBuildInstaller(Installer):
             raise MethodNotSupportedError(f"{SYSTEM} OS not supported")
 
 
+@GitAnnexComponent.register_installer
+class DMGInstaller(Installer):
+    """ Installs a local ``*.dmg`` file """
+
+    NAME = "dmg"
+
+    OPTIONS = [
+        Option(
+            "--path",
+            converter=Path,
+            metavar="PATH",
+            help="Path to local `*.dmg` to install",
+        ),
+    ]
+
+    PACKAGES = {
+        "git-annex": ("git-annex", ["git-annex"]),
+    }
+
+    def install_package(
+        self,
+        package: str,
+        path: Optional[Path] = None,
+        **kwargs: Any,
+    ) -> Path:
+        log.info("Installing %s via dmg", package)
+        if path is None:
+            raise RuntimeError("dmg method requires path")
+        log.info("Path: %s", path)
+        if kwargs:
+            log.warning("Ignoring extra installer arguments: %r", kwargs)
+        binpath = install_git_annex_dmg(path, self.manager)
+        log.debug("Installed program directory: %s", binpath)
+        return binpath
+
+    def assert_supported_system(self) -> None:
+        if not ON_MACOS:
+            raise MethodNotSupportedError(f"{SYSTEM} OS not supported")
+
+
 class MethodNotSupportedError(Exception):
     """
     Raised when an installer's `install()` method is called on an unsupported
@@ -1874,6 +2048,31 @@ def install_git_annex_dmg(
     return annex_bin
 
 
+def install_deb(
+    debpath: Union[str, os.PathLike],
+    manager: DataladInstaller,
+    bin_path: Path,
+    install_dir: Optional[Path] = None,
+    extra_args: Optional[List[str]] = None,
+) -> Path:
+    cmd: List[Union[str, os.PathLike]] = ["dpkg"]
+    if extra_args is not None:
+        cmd.extend(extra_args)
+    if install_dir is None:
+        cmd.append("-i")
+        cmd.append(debpath)
+        manager.sudo(*cmd)
+        return Path("/usr/bin")
+    else:
+        install_dir.mkdir(parents=True, exist_ok=True)
+        cmd.append("-x")
+        cmd.append(debpath)
+        cmd.append(install_dir)
+        runcmd(*cmd)
+        manager.addpath(install_dir / bin_path)
+        return install_dir / bin_path
+
+
 def ask(prompt: str, choices: List[str]) -> str:
     full_prompt = f"{prompt} [{'/'.join(choices)}] "
     while True:
@@ -1896,6 +2095,43 @@ def get_version_codename() -> str:
         return fp.read().partition("/")[0]
 
 
+def parse_header_links(links_header: str) -> Dict[str, Dict[str, str]]:
+    """
+    Parse a "Link" header from an HTTP response into a `dict` of the form::
+
+        {"next": {"url": "...", "rel": "next"}, "last": { ... }}
+    """
+    # <https://git.io/JcYZi>
+    links: Dict[str, Dict[str, str]] = {}
+    replace_chars = " '\""
+    value = links_header.strip(replace_chars)
+    if not value:
+        return links
+    for val in re.split(r", *<", value):
+        try:
+            url, params = val.split(";", 1)
+        except ValueError:
+            url, params = val, ""
+        link: Dict[str, str] = {"url": url.strip("<> '\"")}
+        for param in params.split(";"):
+            try:
+                key, value = param.split("=")
+            except ValueError:
+                break
+            link[key.strip(replace_chars)] = value.strip(replace_chars)
+        key = link.get("rel") or link.get("url")
+        assert key is not None
+        links[key] = link
+    return links
+
+
+def untmppath(path: Path) -> Path:
+    if "{tmpdir}" in str(path):
+        return Path(str(path).format(tmpdir=mktempdir("dl-")))
+    else:
+        return path
+
+
 def main(argv: Optional[List[str]] = None) -> int:
     with DataladInstaller() as manager:
         return manager.main(argv)
